#if defined(WIN32)
#include <tchar.h>
#endif

#include <cstdlib>
#include <utility>
#include <iostream>
#include <boost/noncopyable.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/shared_ptr.hpp>

namespace intel_shared_ptr_to_bool_test {

template <typename Value>
class sp_intrusive_list : private boost::noncopyable
{
public:
  typedef Value  value_type;
  typedef Value* pointer;
  typedef Value& reference;
  typedef boost::weak_ptr<Value>   weak_pointer;
  typedef boost::shared_ptr<Value> shared_pointer;

  /// Required hook for items of the list.
  class base_hook : private boost::noncopyable
  {
  private:
    friend class sp_intrusive_list<value_type>;
    weak_pointer   prev_;
    shared_pointer next_;
  }; // class base_hook

  /// Never throws
  sp_intrusive_list()
    : size_(0)
  {
  }

  ~sp_intrusive_list()
  {
    clear();
  }

  /// Never throws
  shared_pointer front() const
  {
    return front_;
  }

  /// Never throws
  static shared_pointer prev(const shared_pointer& value)
  {
    BOOST_ASSERT(value);
    return get_hook(*value).prev_.lock();
  }

  /// Never throws
  static shared_pointer next(const shared_pointer& value)
  {
    BOOST_ASSERT(value);
    return get_hook(*value).next_;
  }

  /// Never throws
  void push_front(const shared_pointer& value)
  {
    BOOST_ASSERT(value);

    base_hook& value_hook = get_hook(*value);

    BOOST_ASSERT(!value_hook.prev_.lock() && !value_hook.next_);

    value_hook.next_ = front_;
    if (front_)
    {
      base_hook& front_hook = get_hook(*front_);
      front_hook.prev_ = value;
    }
    front_ = value;
    ++size_;
  }

  void erase(const shared_pointer& value)
  {
    BOOST_ASSERT(value);

    base_hook& value_hook = get_hook(*value);
    if (value == front_)
    {
      front_ = value_hook.next_;
    }
    const shared_pointer prev = value_hook.prev_.lock();
    if (prev)
    {
      base_hook& prev_hook = get_hook(*prev);
      prev_hook.next_ = value_hook.next_;
    }
    if (value_hook.next_)
    {
      base_hook& next_hook = get_hook(*value_hook.next_);
      next_hook.prev_ = value_hook.prev_;
    }
    value_hook.prev_.reset();
    value_hook.next_.reset();
    --size_;

    BOOST_ASSERT(!value_hook.prev_.lock() && !value_hook.next_);
  }

  void clear()
  {
    // We don't want to have recusrive calls of wrapped_session's destructor
    // because the deep of such recursion may be equal to the size of list.
    // The last can be too great for the stack.
    while (front_)
    {
      base_hook& front_hook = get_hook(*front_);
#if defined(MA_HAS_RVALUE_REFS)
      shared_pointer tmp = std::move(front_hook.next_);
#else
      shared_pointer tmp = front_hook.next_;
      front_hook.next_.reset();
#endif
      front_hook.prev_.reset();
      front_.swap(tmp);
    }
    size_ = 0;
  }

  std::size_t size() const
  {
    return size_;
  }

  bool empty() const
  {
    return 0 == size_;
  }

private:
  static base_hook& get_hook(reference value)
  {
    return static_cast<base_hook&>(value);
  }

  std::size_t size_;
  shared_pointer front_;
}; // class sp_intrusive_list

class A;
typedef boost::shared_ptr<A> A_ptr;

class A 
  : private boost::noncopyable
  , public sp_intrusive_list<A>::base_hook
{
}; // class A

void run_test()
{
  sp_intrusive_list<A> list;  
  A_ptr a = list.front();
  if (a)
  {
    std::cout << "intel_shared_ptr_to_bool_test failed" << std::endl;
  }
  else
  {
    std::cout << "intel_shared_ptr_to_bool_test passed" << std::endl;
  }  
}

} // namespace intel_shared_ptr_to_bool_test

#if defined(WIN32)
int _tmain(int /*argc*/, _TCHAR* /*argv*/[])
#else
int main(int /*argc*/, char* /*argv*/[])
#endif
{
  intel_shared_ptr_to_bool_test::run_test();
  return EXIT_SUCCESS;
}
